import base64
from datetime import datetime
import hashlib
import re
import requests
from Crypto.PublicKey.RSA import construct
from Crypto.Cipher import PKCS1_v1_5
import binascii
import string, random
import time

class TplinkDeviceScanner:
    """Abstract parent class for all TPLink device scanners."""

    def __init__(self, host, username, password):
        """Generic scanner initialization."""
        self.parse_macs_hyphens = re.compile('[0-9A-F]{2}-[0-9A-F]{2}-' +
                                             '[0-9A-F]{2}-[0-9A-F]{2}-' +
                                             '[0-9A-F]{2}-[0-9A-F]{2}')
        self.parse_macs_colons = re.compile('[0-9A-F]{2}:[0-9A-F]{2}:' +
                                            '[0-9A-F]{2}:[0-9A-F]{2}:' +
                                            '[0-9A-F]{2}:[0-9A-F]{2}')

        self.host = host
        self.username = username
        self.password = password

        self.last_results = {}
        self.success_init = self._update_info()

    def scan_devices(self):
        """Scan for new devices and return a list with found device IDs."""
        self._update_info()
        return self.last_results

    def get_base64_cookie_string(self):
        """Encode Base 64 authentication string for some scanners."""
        username_password = '{}:{}'.format(self.username, self.password)
        b64_encoded_username_password = base64.b64encode(
            username_password.encode('ascii')
        ).decode('ascii')
        return 'Authorization=Basic {}'.format(b64_encoded_username_password)

class VR600TplinkDeviceScanner(TplinkDeviceScanner):
    """This class queries an Archer VR600 router with TP-Link firmware."""

    def __init__(self, host, username, password):
        """Initialize the scanner."""
        self.pubkey = ''
        self.jsessionId = ''
        self.token = ''
        super(VR600TplinkDeviceScanner, self).__init__(host, username, password)

    def _get_pub_key(self):
        # Get the modulu and exponent from the router
        url = 'http://{}/cgi/getParm'.format(self.host)
        referer = 'http://{}'.format(self.host)
        response = requests.post(url, headers={ 'REFERER': referer})
        if not response.status_code == 200:
            return False

        ee = self._get_field_from_router_response(response.text, 'ee')
        nn = self._get_field_from_router_response(response.text, 'nn')

        try:
            e = int(ee, 16)
            n = int(nn, 16)  #snipped for brevity
        except ValueError:
            return False

        pubkey = construct((n, e))
        self.pubkey = PKCS1_v1_5.new(pubkey)
        print("Pub Key: " + str(pubkey))

        return True

    def _get_field_from_router_response(self, rText, key):
        lines = rText.split('\n')
        for line in lines:
            if (line.startswith('var '+ key)):
                # var ee="010101" => "010101"
                return line.split('=\"')[1].split('\";')[0]
        return ''

    def _get_jsession_id(self):
        username = self.username.encode('utf-8')
        password = self.password.encode('utf-8')

        b64pass = base64.b64encode(password)
        encryptedUsername = self.pubkey.encrypt(username)
        encryptedPassword = self.pubkey.encrypt(b64pass)

        print("Encrypted Pass: " + encryptedPassword)

        base16username = base64.b16encode(encryptedUsername).decode('utf-8').lower()
        base16password = base64.b16encode(encryptedPassword).decode('utf-8').lower()

        referer = 'http://{}'.format(self.host)
        url = 'http://{}/cgi/login?UserName={}&Passwd={}&Action=1&LoginStatus=0'.format(self.host, base16username, base16password)

        headers= { 'referer': referer }

        response = requests.post(url, headers=headers)

        if not response.status_code == 200:
            print("Error %s from router", response)
            return False

        self.jsessionId = dict(response.cookies)['JSESSIONID']
        print(response.cookies)
        return True

    def _get_random_alphaNumeric_string(self, stringLength=15):
        lettersAndDigits = string.ascii_letters + string.digits
        return ''.join((random.choice(lettersAndDigits) for i in range(stringLength)))

    def _get_token(self):
        url = 'http://{}/'.format(self.host)
        referer = 'http://{}'.format(self.host)

        headers= {
            'referer': referer,
            'cookie': 'loginErrorShow=1; JSESSIONID='+self.jsessionId,
            }

        cookies = {'JSESSIONID': self.jsessionId}
        print(cookies)
        response = requests.get(url, headers=headers, cookies=cookies)

        if not response.status_code == 200:
            print("Error %s from router", response)
            return False

        split = response.text.index('var token=') + len('var token=\"') 
        token = response.text[split:split+30]

        self.token = token
        return True

    def _get_auth_tokens(self):
        """Retrieve auth tokens from the router."""

        print("Retrieving PublicKey...")
        pubkey = self._get_pub_key()
        if not pubkey:
            print("Pub key fetch failed")
            return False

        print("Retrieving JSessionID...")
        jsessionId = self._get_jsession_id()
        if not jsessionId:
            print("jsessionId fetch failed")
            return False

        print("Retrieving Token...")
        token = self._get_token()
        if not token:
            print("token fetch failed")
            return False

        return True

    def _get_mac_results(self):
        referer = 'http://{}'.format(self.host)
        headers= {
            'TokenID': self.token,
            'referer': referer,
            'cookie': 'JSESSIONID=' + self.jsessionId
            }

        mac_results = []

        # Check both the 2.4GHz and 5GHz client lists.
        for clients_frequency in ('1', '2'):

            # Refresh associated clients.
            page = requests.post(
                'http://{}/cgi?7'.format(self.host),
                headers=headers,
                data=(
                    '[ACT_WLAN_UPDATE_ASSOC#1,{},0,0,0,0#0,0,0,0,0,0]0,0\r\n'
                    ).format(clients_frequency),
                timeout=4)
            if not page.status_code == 200:
                print("Error %s from router", page.status_code)
                return False

            # Retrieve associated clients.
            page = requests.post(
                'http://{}/cgi?6'.format(self.host),
                headers=headers,
                data=(
                    '[LAN_WLAN_ASSOC_DEV#0,0,0,0,0,0#1,{},0,0,0,0]0,1\r\n'
                    'AssociatedDeviceMACAddress\r\n'
                    ).format(clients_frequency),
                timeout=4)
            if not page.status_code == 200:
                print("Error %s from router", page.status_code)
                return False

            mac_results.extend(self.parse_macs_colons.findall(page.text))
    
        self.last_results = mac_results
        return True

    def _update_info(self):
        """Ensure the information from the TP-Link router is up to date.
        Return boolean if scanning successful.
        """
        print("[VR600] Loading wireless clients...")

        if (self.jsessionId == '') or (self.token == ''):
            gotToken = self._get_auth_tokens()
            if not gotToken:
                # Retry
                print("Failed to get AuthTokens. Retrying in 3 secs.")
                time.sleep(3)
                gotToken = self._get_auth_tokens()
        else:
            gotToken = True

        if not gotToken:
            """ In case of failure - force re-login """
            self.jsessionId = ''
            self.token = ''
            return False

        macResults = self._get_mac_results()
        if not macResults:
            """ In case of failure - force re-login """
            self.jsessionId = ''
            self.token = ''
            return False
        return True

router = VR600TplinkDeviceScanner('192.168.0.1','username','yourpassword')